import logging
from typing import Optional

from pymint.antlrgen.mintListener import mintListener
from pymint.antlrgen.mintParser import mintParser
from pymint.mintdevice import MINTDevice
from pymint.mintlayer import MINTLayer, MINTLayerType
from pymint.minttarget import MINTTarget


class MINTCompiler(mintListener):
    """Primary ANTLR listener class for the compiler"""

    def __init__(self):
        super().__init__()
        self.current_device: Optional[MINTDevice] = None
        self.current_block_id = 0
        self.current_layer_id = 0
        self.flow_layer_count = 0
        self.control_layer_count = 0
        self.integration_layer_count = 0
        self.current_entity: Optional[str]
        self.current_params = {}
        self._current_layer: Optional[MINTLayer] = None

    def enterNetlist(self, ctx: mintParser.NetlistContext):
        self.current_device = MINTDevice("DEFAULT_NAME")

    def exitNetlist(self, ctx: mintParser.NetlistContext):
        if self.current_device is None:
            raise Exception("Could not find the device")
        self.current_device.generate_network()

    def enterHeader(self, ctx: mintParser.HeaderContext):
        if ctx.device_name is None:
            raise Exception("Could not find Device Name")
        self.current_device.name = ctx.device_name.text

    def exitLayerBlock(self, ctx: mintParser.LayerBlockContext):
        # Increement teh layer block
        self.current_block_id += 1

    def enterFlowBlock(self, ctx: mintParser.FlowBlockContext):
        layer = self.current_device.create_mint_layer(
            str(self.current_layer_id),
            str(self.flow_layer_count),
            str(self.current_block_id),
            MINTLayerType.FLOW,
        )
        self._current_layer = layer
        self.flow_layer_count += 1
        self.current_layer_id += 1

    def enterControlBlock(self, ctx: mintParser.ControlBlockContext):
        layer = self.current_device.create_mint_layer(
            str(self.current_layer_id),
            str(self.control_layer_count),
            str(self.current_block_id),
            MINTLayerType.CONTROL,
        )
        self._current_layer = layer
        self.control_layer_count += 1
        self.current_layer_id += 1

    def enterIntegrationBlock(self, ctx: mintParser.IntegrationBlockContext):
        layer = self.current_device.create_mint_layer(
            str(self.current_layer_id),
            str(self.integration_layer_count),
            str(self.current_block_id),
            MINTLayerType.INTEGRATION,
        )
        self._current_layer = layer
        self.integration_layer_count += 1
        self.current_layer_id += 1

    def enterEntity(self, ctx: mintParser.EntityContext):
        self.current_entity = ctx.getText()

    def enterParamsStat(self, ctx: mintParser.ParamsStatContext):
        self.current_params = {}

    def enterIntParam(self, ctx: mintParser.IntParamContext):
        value = ctx.value().getText()
        key = ctx.param_element().getText()
        self.current_params[key] = int(value)

    def enterBoolParam(self, ctx: mintParser.BoolParamContext):
        if ctx.boolvalue.getText() == "YES":
            value = True
        else:
            value = False
        key = ctx.param_element.getText()
        self.current_params[key] = value

    def enterLengthParam(self, ctx: mintParser.LengthParamContext):
        value = float(ctx.value().getText())
        self.current_params["length"] = value

    def enterSpacingParam(self, ctx: mintParser.SpacingParamContext):
        value = float(ctx.value().getText())
        self.current_params["spacing"] = value

    def enterWidthParam(self, ctx: mintParser.WidthParamContext):
        value = ctx.value().getText()
        if ctx.key is None:
            raise AssertionError
        key = ctx.key.text
        if key is None:
            raise Exception("Error in parsing the width parameter")
        if key == "w":
            key = "width"

        self.current_params[key] = int(value)

    def enterFlowStat(self, ctx: mintParser.FlowStatContext):
        self.current_entity = None
        self.current_params = {}

    def enterControlStat(self, ctx: mintParser.ControlStatContext):
        self.current_entity = None
        self.current_params = {}

    def exitPrimitiveStat(self, ctx: mintParser.PrimitiveStatContext):
        entity = self.current_entity
        if entity is None:
            raise Exception("Could not find the technology for the pimitive")

        # Loop for each of the components that need to be created with this param
        for ufname in ctx.ufnames().ufname():
            if self._current_layer is None:
                raise Exception("Current layer is set to None")
            if not (
                self._current_layer is not None and self._current_layer.ID is not None
            ):
                raise AssertionError
            self.current_device.create_mint_component(
                ufname.getText(),
                entity,
                self.current_params,
                [self._current_layer.ID],
            )

    def exitBankDeclStat(self, ctx: mintParser.BankDeclStatContext):
        entity = self.current_entity
        if entity is None:
            raise Exception("Could not find the technology for the primitive")

        # Clean up the constraint specific params
        self._cleanup_BANK_params()

        for ufname in ctx.ufnames().ufname():
            component_name = ufname.getText()
            if self._current_layer.ID is None:
                raise AssertionError
            self.current_device.create_mint_component(
                component_name,
                entity,
                self.current_params,
                [self._current_layer.ID],
            )

    def exitBankGenStat(self, ctx: mintParser.BankGenStatContext):
        entity = self.current_entity
        if entity is None:
            raise Exception("Could not find the technology for the primitive")

        self._cleanup_BANK_params()

        if ctx.dim is None:
            raise AssertionError
        dim = int(ctx.dim.text)

        name = ctx.ufname().getText()

        for i in range(1, dim + 1):
            component_name = name + "_" + str(i)
            if self._current_layer is None:
                raise Exception("Current Layer not Set")
            if not (
                self._current_layer is not None and self._current_layer.ID is not None
            ):
                raise AssertionError
            self.current_device.create_mint_component(
                component_name, entity, self.current_params, [self._current_layer.ID]
            )

    def exitGridDeclStat(self, ctx: mintParser.GridDeclStatContext):
        entity = self.current_entity
        if entity is None:
            raise Exception("Could not find the technology for the primitive")

        self._cleanup_GRID_params()

        for ufname in ctx.ufnames().ufname():
            component_name = ufname.getText()
            if self._current_layer.ID is None:
                raise AssertionError
            self.current_device.create_mint_component(
                component_name,
                entity,
                self.current_params,
                [self._current_layer.ID],
            )

    def exitChannelStat(self, ctx: mintParser.ChannelStatContext):
        entity = self.current_entity
        if entity is None:
            entity = "CHANNEL"
        connection_name = ctx.ufname().getText()

        source_target = ctx.uftarget()[0]
        source_id = source_target.ID().getText()
        if self.current_device.component_exists(source_id) is False:
            raise Exception(
                "Error ! - Could not find the component '{}' in device '{}'".format(
                    source_id, self.current_device.name
                )
            )
        if source_target.INT():
            source_port = source_target.INT().getText()
        else:
            source_port = None

        source_uftarget = MINTTarget(source_id, source_port)

        sink_target = ctx.uftarget()[1]
        sink_id = sink_target.ID().getText()
        if self.current_device.component_exists(sink_id) is False:
            raise Exception(
                "Error ! - Could not find the component '{}' in device '{}'".format(
                    sink_id, self.current_device.name
                )
            )
        if sink_target.INT():
            sink_port = sink_target.INT().getText()
        else:
            sink_port = None

        sink_uftarget = MINTTarget(sink_id, sink_port)

        self._cleanup_CHANNEL_params()

        # Create a connection between the different components in the device
        if not (self._current_layer is not None and self._current_layer.ID is not None):
            raise AssertionError
        self.current_device.create_mint_connection(
            connection_name,
            entity,
            self.current_params,
            source_uftarget,
            [sink_uftarget],
            self._current_layer.ID,
        )

    def exitNetStat(self, ctx: mintParser.NetStatContext):
        entity = self.current_entity
        if entity is None:
            entity = "NET"

        connection_name = ctx.ufname().getText()

        source_target = ctx.uftarget()
        source_id = source_target.ID().getText()
        if source_target.INT():
            source_port = source_target.INT().getText()
        else:
            source_port = None

        source_uftarget = MINTTarget(source_id, source_port)

        sink_uftargets = []

        for sink_target in ctx.uftargets().uftarget():
            sink_id = sink_target.ID().getText()
            if sink_target.INT():
                sink_port = sink_target.INT().getText()
            else:
                sink_port = None

            sink_uftargets.append(MINTTarget(sink_id, sink_port))
        if not (self._current_layer is not None and self._current_layer.ID is not None):
            raise AssertionError
        self.current_device.create_mint_connection(
            connection_name,
            entity,
            self.current_params,
            source_uftarget,
            sink_uftargets,
            self._current_layer.ID,
        )

    def exitSpanStat(self, ctx: mintParser.SpanStatContext):
        entity = self.current_entity
        if entity is None:
            raise Exception("Could not find the technology for the pimitive")

        # pipe in the in / out values to params for sizing
        if ctx.indim is not None:
            in_value = int(ctx.indim.text)
            self.current_params["in"] = in_value
        if ctx.outdim is not None:
            out_value = int(ctx.outdim.text)
            self.current_params["out"] = out_value

        # Loop for each of the components that need to be created with this param
        for ufname in ctx.ufnames().ufname():
            if self._current_layer.ID is None:
                raise AssertionError
            self.current_device.create_mint_component(
                ufname.getText(),
                entity,
                self.current_params,
                [self._current_layer.ID],
            )

    def exitNodeStat(self, ctx: mintParser.NodeStatContext):
        entity = "NODE"

        # Loop for each of the components that need to be created with this param
        if not (self._current_layer is not None and self._current_layer.ID is not None):
            raise AssertionError
        for ufname in ctx.ufnames().ufname():
            self.current_device.create_mint_component(
                ufname.getText(),
                entity,
                self.current_params,
                [self._current_layer.ID],
            )

    def exitValveStat(self, ctx: mintParser.ValveStatContext):
        entity = self.current_entity
        if entity is None:
            logging.error("Could not find entitry information for valve")
            raise Exception("Could not find entitry information for valve")
        valve_name = ctx.ufname()[0].getText()
        if not (self._current_layer is not None and self._current_layer.ID is not None):
            raise AssertionError
        valve_component = self.current_device.create_mint_component(
            valve_name,
            entity,
            self.current_params,
            [self._current_layer.ID],
        )
        connection_name = ctx.ufname()[1].getText()
        valve_connection = self.current_device.get_connection(connection_name)
        if valve_connection is None:
            raise Exception(
                "Error: Could not find connection '{}' in device '{}'".format(
                    connection_name, self.current_device.name
                )
            )

        self.current_device.map_valve(valve_component, valve_connection)

    def enterViaStat(self, ctx: mintParser.ViaStatContext):
        for ufname in ctx.ufnames().ufname():
            self.current_device.add_via(ufname.getText(), [])

    def enterTerminalStat(self, ctx: mintParser.TerminalStatContext):
        terminal_name = ctx.ufname().getText()
        pin_number = int(ctx.INT.getText())
        if not (self._current_layer is not None and self._current_layer.ID is not None):
            raise AssertionError
        self.current_device.add_terminal(
            terminal_name,
            pin_number,
            self._current_layer.ID,
        )

    def _cleanup_BANK_params(self):
        if "spacing" in self.current_params.keys():
            del self.current_params["spacing"]

    def _cleanup_GRID_params(self):
        if "horizontalSpacing" in self.current_params.keys():
            del self.current_params["horizontalSpacing"]
        if "verticalSpacing" in self.current_params.keys():
            del self.current_params["verticalSpacing"]

    def _cleanup_CHANNEL_params(self):
        if "length" in self.current_params.keys():
            del self.current_params["length"]



import argparse
import json
import os
from pathlib import Path

from pymint import MINTDevice


def main():
    """Scaffhold for running the command line interface"""
    # do the arg parsing here
    parser = argparse.ArgumentParser()

    parser.add_argument("input", help="This is the file thats used as the input ")
    parser.add_argument(
        "--outpath", type=str, default="out/", help="This is the output directory"
    )
    parser.add_argument(
        "--skip-constraints",
        action="store_true",
        help=(
            "Sets the flag to only convert the design and not try to generate the"
            " constraints"
        ),
    )

    args = parser.parse_args()
    OUTPUT_DIR = Path(args.outpath).resolve()
    file_path = str(Path(args.input).resolve())

    current_device = MINTDevice.from_mint_file(file_path, skip_constraints=True)

    tt = os.path.join(OUTPUT_DIR, "{}.json".format(current_device.name))
    with open(tt, "w") as f:
        json.dump(current_device.to_parchmint_v1(), f)


if __name__ == "__main__":
    main()



# Generated from ./mint.g4 by ANTLR 4.9.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .mintParser import mintParser
else:
    from mintParser import mintParser

# This class defines a complete listener for a parse tree produced by mintParser.
class mintListener(ParseTreeListener):

    # Enter a parse tree produced by mintParser#netlist.
    def enterNetlist(self, ctx:mintParser.NetlistContext):
        pass

    # Exit a parse tree produced by mintParser#netlist.
    def exitNetlist(self, ctx:mintParser.NetlistContext):
        pass


    # Enter a parse tree produced by mintParser#importBlock.
    def enterImportBlock(self, ctx:mintParser.ImportBlockContext):
        pass

    # Exit a parse tree produced by mintParser#importBlock.
    def exitImportBlock(self, ctx:mintParser.ImportBlockContext):
        pass


    # Enter a parse tree produced by mintParser#importStat.
    def enterImportStat(self, ctx:mintParser.ImportStatContext):
        pass

    # Exit a parse tree produced by mintParser#importStat.
    def exitImportStat(self, ctx:mintParser.ImportStatContext):
        pass


    # Enter a parse tree produced by mintParser#header.
    def enterHeader(self, ctx:mintParser.HeaderContext):
        pass

    # Exit a parse tree produced by mintParser#header.
    def exitHeader(self, ctx:mintParser.HeaderContext):
        pass


    # Enter a parse tree produced by mintParser#ufmoduleBlock.
    def enterUfmoduleBlock(self, ctx:mintParser.UfmoduleBlockContext):
        pass

    # Exit a parse tree produced by mintParser#ufmoduleBlock.
    def exitUfmoduleBlock(self, ctx:mintParser.UfmoduleBlockContext):
        pass


    # Enter a parse tree produced by mintParser#globalStats.
    def enterGlobalStats(self, ctx:mintParser.GlobalStatsContext):
        pass

    # Exit a parse tree produced by mintParser#globalStats.
    def exitGlobalStats(self, ctx:mintParser.GlobalStatsContext):
        pass


    # Enter a parse tree produced by mintParser#ufmoduleStat.
    def enterUfmoduleStat(self, ctx:mintParser.UfmoduleStatContext):
        pass

    # Exit a parse tree produced by mintParser#ufmoduleStat.
    def exitUfmoduleStat(self, ctx:mintParser.UfmoduleStatContext):
        pass


    # Enter a parse tree produced by mintParser#layerBlocks.
    def enterLayerBlocks(self, ctx:mintParser.LayerBlocksContext):
        pass

    # Exit a parse tree produced by mintParser#layerBlocks.
    def exitLayerBlocks(self, ctx:mintParser.LayerBlocksContext):
        pass


    # Enter a parse tree produced by mintParser#layerBlock.
    def enterLayerBlock(self, ctx:mintParser.LayerBlockContext):
        pass

    # Exit a parse tree produced by mintParser#layerBlock.
    def exitLayerBlock(self, ctx:mintParser.LayerBlockContext):
        pass


    # Enter a parse tree produced by mintParser#flowBlock.
    def enterFlowBlock(self, ctx:mintParser.FlowBlockContext):
        pass

    # Exit a parse tree produced by mintParser#flowBlock.
    def exitFlowBlock(self, ctx:mintParser.FlowBlockContext):
        pass


    # Enter a parse tree produced by mintParser#controlBlock.
    def enterControlBlock(self, ctx:mintParser.ControlBlockContext):
        pass

    # Exit a parse tree produced by mintParser#controlBlock.
    def exitControlBlock(self, ctx:mintParser.ControlBlockContext):
        pass


    # Enter a parse tree produced by mintParser#integrationBlock.
    def enterIntegrationBlock(self, ctx:mintParser.IntegrationBlockContext):
        pass

    # Exit a parse tree produced by mintParser#integrationBlock.
    def exitIntegrationBlock(self, ctx:mintParser.IntegrationBlockContext):
        pass


    # Enter a parse tree produced by mintParser#flowStat.
    def enterFlowStat(self, ctx:mintParser.FlowStatContext):
        pass

    # Exit a parse tree produced by mintParser#flowStat.
    def exitFlowStat(self, ctx:mintParser.FlowStatContext):
        pass


    # Enter a parse tree produced by mintParser#controlStat.
    def enterControlStat(self, ctx:mintParser.ControlStatContext):
        pass

    # Exit a parse tree produced by mintParser#controlStat.
    def exitControlStat(self, ctx:mintParser.ControlStatContext):
        pass


    # Enter a parse tree produced by mintParser#integrationStat.
    def enterIntegrationStat(self, ctx:mintParser.IntegrationStatContext):
        pass

    # Exit a parse tree produced by mintParser#integrationStat.
    def exitIntegrationStat(self, ctx:mintParser.IntegrationStatContext):
        pass


    # Enter a parse tree produced by mintParser#primitiveStat.
    def enterPrimitiveStat(self, ctx:mintParser.PrimitiveStatContext):
        pass

    # Exit a parse tree produced by mintParser#primitiveStat.
    def exitPrimitiveStat(self, ctx:mintParser.PrimitiveStatContext):
        pass


    # Enter a parse tree produced by mintParser#bankDeclStat.
    def enterBankDeclStat(self, ctx:mintParser.BankDeclStatContext):
        pass

    # Exit a parse tree produced by mintParser#bankDeclStat.
    def exitBankDeclStat(self, ctx:mintParser.BankDeclStatContext):
        pass


    # Enter a parse tree produced by mintParser#bankGenStat.
    def enterBankGenStat(self, ctx:mintParser.BankGenStatContext):
        pass

    # Exit a parse tree produced by mintParser#bankGenStat.
    def exitBankGenStat(self, ctx:mintParser.BankGenStatContext):
        pass


    # Enter a parse tree produced by mintParser#bankStat.
    def enterBankStat(self, ctx:mintParser.BankStatContext):
        pass

    # Exit a parse tree produced by mintParser#bankStat.
    def exitBankStat(self, ctx:mintParser.BankStatContext):
        pass


    # Enter a parse tree produced by mintParser#gridGenStat.
    def enterGridGenStat(self, ctx:mintParser.GridGenStatContext):
        pass

    # Exit a parse tree produced by mintParser#gridGenStat.
    def exitGridGenStat(self, ctx:mintParser.GridGenStatContext):
        pass


    # Enter a parse tree produced by mintParser#gridDeclStat.
    def enterGridDeclStat(self, ctx:mintParser.GridDeclStatContext):
        pass

    # Exit a parse tree produced by mintParser#gridDeclStat.
    def exitGridDeclStat(self, ctx:mintParser.GridDeclStatContext):
        pass


    # Enter a parse tree produced by mintParser#gridStat.
    def enterGridStat(self, ctx:mintParser.GridStatContext):
        pass

    # Exit a parse tree produced by mintParser#gridStat.
    def exitGridStat(self, ctx:mintParser.GridStatContext):
        pass


    # Enter a parse tree produced by mintParser#spanStat.
    def enterSpanStat(self, ctx:mintParser.SpanStatContext):
        pass

    # Exit a parse tree produced by mintParser#spanStat.
    def exitSpanStat(self, ctx:mintParser.SpanStatContext):
        pass


    # Enter a parse tree produced by mintParser#valveStat.
    def enterValveStat(self, ctx:mintParser.ValveStatContext):
        pass

    # Exit a parse tree produced by mintParser#valveStat.
    def exitValveStat(self, ctx:mintParser.ValveStatContext):
        pass


    # Enter a parse tree produced by mintParser#nodeStat.
    def enterNodeStat(self, ctx:mintParser.NodeStatContext):
        pass

    # Exit a parse tree produced by mintParser#nodeStat.
    def exitNodeStat(self, ctx:mintParser.NodeStatContext):
        pass


    # Enter a parse tree produced by mintParser#viaStat.
    def enterViaStat(self, ctx:mintParser.ViaStatContext):
        pass

    # Exit a parse tree produced by mintParser#viaStat.
    def exitViaStat(self, ctx:mintParser.ViaStatContext):
        pass


    # Enter a parse tree produced by mintParser#terminalStat.
    def enterTerminalStat(self, ctx:mintParser.TerminalStatContext):
        pass

    # Exit a parse tree produced by mintParser#terminalStat.
    def exitTerminalStat(self, ctx:mintParser.TerminalStatContext):
        pass


    # Enter a parse tree produced by mintParser#channelStat.
    def enterChannelStat(self, ctx:mintParser.ChannelStatContext):
        pass

    # Exit a parse tree produced by mintParser#channelStat.
    def exitChannelStat(self, ctx:mintParser.ChannelStatContext):
        pass


    # Enter a parse tree produced by mintParser#netStat.
    def enterNetStat(self, ctx:mintParser.NetStatContext):
        pass

    # Exit a parse tree produced by mintParser#netStat.
    def exitNetStat(self, ctx:mintParser.NetStatContext):
        pass


    # Enter a parse tree produced by mintParser#entity.
    def enterEntity(self, ctx:mintParser.EntityContext):
        pass

    # Exit a parse tree produced by mintParser#entity.
    def exitEntity(self, ctx:mintParser.EntityContext):
        pass


    # Enter a parse tree produced by mintParser#entity_element.
    def enterEntity_element(self, ctx:mintParser.Entity_elementContext):
        pass

    # Exit a parse tree produced by mintParser#entity_element.
    def exitEntity_element(self, ctx:mintParser.Entity_elementContext):
        pass


    # Enter a parse tree produced by mintParser#paramsStat.
    def enterParamsStat(self, ctx:mintParser.ParamsStatContext):
        pass

    # Exit a parse tree produced by mintParser#paramsStat.
    def exitParamsStat(self, ctx:mintParser.ParamsStatContext):
        pass


    # Enter a parse tree produced by mintParser#statTerminaion.
    def enterStatTerminaion(self, ctx:mintParser.StatTerminaionContext):
        pass

    # Exit a parse tree produced by mintParser#statTerminaion.
    def exitStatTerminaion(self, ctx:mintParser.StatTerminaionContext):
        pass


    # Enter a parse tree produced by mintParser#connectionParamStat.
    def enterConnectionParamStat(self, ctx:mintParser.ConnectionParamStatContext):
        pass

    # Exit a parse tree produced by mintParser#connectionParamStat.
    def exitConnectionParamStat(self, ctx:mintParser.ConnectionParamStatContext):
        pass


    # Enter a parse tree produced by mintParser#paramStat.
    def enterParamStat(self, ctx:mintParser.ParamStatContext):
        pass

    # Exit a parse tree produced by mintParser#paramStat.
    def exitParamStat(self, ctx:mintParser.ParamStatContext):
        pass


    # Enter a parse tree produced by mintParser#constraintParams.
    def enterConstraintParams(self, ctx:mintParser.ConstraintParamsContext):
        pass

    # Exit a parse tree produced by mintParser#constraintParams.
    def exitConstraintParams(self, ctx:mintParser.ConstraintParamsContext):
        pass


    # Enter a parse tree produced by mintParser#spacingParam.
    def enterSpacingParam(self, ctx:mintParser.SpacingParamContext):
        pass

    # Exit a parse tree produced by mintParser#spacingParam.
    def exitSpacingParam(self, ctx:mintParser.SpacingParamContext):
        pass


    # Enter a parse tree produced by mintParser#directionParam.
    def enterDirectionParam(self, ctx:mintParser.DirectionParamContext):
        pass

    # Exit a parse tree produced by mintParser#directionParam.
    def exitDirectionParam(self, ctx:mintParser.DirectionParamContext):
        pass


    # Enter a parse tree produced by mintParser#param_element.
    def enterParam_element(self, ctx:mintParser.Param_elementContext):
        pass

    # Exit a parse tree produced by mintParser#param_element.
    def exitParam_element(self, ctx:mintParser.Param_elementContext):
        pass


    # Enter a parse tree produced by mintParser#intParam.
    def enterIntParam(self, ctx:mintParser.IntParamContext):
        pass

    # Exit a parse tree produced by mintParser#intParam.
    def exitIntParam(self, ctx:mintParser.IntParamContext):
        pass


    # Enter a parse tree produced by mintParser#boolParam.
    def enterBoolParam(self, ctx:mintParser.BoolParamContext):
        pass

    # Exit a parse tree produced by mintParser#boolParam.
    def exitBoolParam(self, ctx:mintParser.BoolParamContext):
        pass


    # Enter a parse tree produced by mintParser#widthParam.
    def enterWidthParam(self, ctx:mintParser.WidthParamContext):
        pass

    # Exit a parse tree produced by mintParser#widthParam.
    def exitWidthParam(self, ctx:mintParser.WidthParamContext):
        pass


    # Enter a parse tree produced by mintParser#verticalSpacingParam.
    def enterVerticalSpacingParam(self, ctx:mintParser.VerticalSpacingParamContext):
        pass

    # Exit a parse tree produced by mintParser#verticalSpacingParam.
    def exitVerticalSpacingParam(self, ctx:mintParser.VerticalSpacingParamContext):
        pass


    # Enter a parse tree produced by mintParser#horizontalSpacingParam.
    def enterHorizontalSpacingParam(self, ctx:mintParser.HorizontalSpacingParamContext):
        pass

    # Exit a parse tree produced by mintParser#horizontalSpacingParam.
    def exitHorizontalSpacingParam(self, ctx:mintParser.HorizontalSpacingParamContext):
        pass


    # Enter a parse tree produced by mintParser#rotationParam.
    def enterRotationParam(self, ctx:mintParser.RotationParamContext):
        pass

    # Exit a parse tree produced by mintParser#rotationParam.
    def exitRotationParam(self, ctx:mintParser.RotationParamContext):
        pass


    # Enter a parse tree produced by mintParser#lengthParam.
    def enterLengthParam(self, ctx:mintParser.LengthParamContext):
        pass

    # Exit a parse tree produced by mintParser#lengthParam.
    def exitLengthParam(self, ctx:mintParser.LengthParamContext):
        pass


    # Enter a parse tree produced by mintParser#ufmodulename.
    def enterUfmodulename(self, ctx:mintParser.UfmodulenameContext):
        pass

    # Exit a parse tree produced by mintParser#ufmodulename.
    def exitUfmodulename(self, ctx:mintParser.UfmodulenameContext):
        pass


    # Enter a parse tree produced by mintParser#ufterminal.
    def enterUfterminal(self, ctx:mintParser.UfterminalContext):
        pass

    # Exit a parse tree produced by mintParser#ufterminal.
    def exitUfterminal(self, ctx:mintParser.UfterminalContext):
        pass


    # Enter a parse tree produced by mintParser#uftargets.
    def enterUftargets(self, ctx:mintParser.UftargetsContext):
        pass

    # Exit a parse tree produced by mintParser#uftargets.
    def exitUftargets(self, ctx:mintParser.UftargetsContext):
        pass


    # Enter a parse tree produced by mintParser#uftarget.
    def enterUftarget(self, ctx:mintParser.UftargetContext):
        pass

    # Exit a parse tree produced by mintParser#uftarget.
    def exitUftarget(self, ctx:mintParser.UftargetContext):
        pass


    # Enter a parse tree produced by mintParser#ufname.
    def enterUfname(self, ctx:mintParser.UfnameContext):
        pass

    # Exit a parse tree produced by mintParser#ufname.
    def exitUfname(self, ctx:mintParser.UfnameContext):
        pass


    # Enter a parse tree produced by mintParser#ufnames.
    def enterUfnames(self, ctx:mintParser.UfnamesContext):
        pass

    # Exit a parse tree produced by mintParser#ufnames.
    def exitUfnames(self, ctx:mintParser.UfnamesContext):
        pass


    # Enter a parse tree produced by mintParser#value.
    def enterValue(self, ctx:mintParser.ValueContext):
        pass

    # Exit a parse tree produced by mintParser#value.
    def exitValue(self, ctx:mintParser.ValueContext):
        pass


    # Enter a parse tree produced by mintParser#boolvalue.
    def enterBoolvalue(self, ctx:mintParser.BoolvalueContext):
        pass

    # Exit a parse tree produced by mintParser#boolvalue.
    def exitBoolvalue(self, ctx:mintParser.BoolvalueContext):
        pass


    # Enter a parse tree produced by mintParser#positionConstraintStat.
    def enterPositionConstraintStat(self, ctx:mintParser.PositionConstraintStatContext):
        pass

    # Exit a parse tree produced by mintParser#positionConstraintStat.
    def exitPositionConstraintStat(self, ctx:mintParser.PositionConstraintStatContext):
        pass


    # Enter a parse tree produced by mintParser#setCoordinate.
    def enterSetCoordinate(self, ctx:mintParser.SetCoordinateContext):
        pass

    # Exit a parse tree produced by mintParser#setCoordinate.
    def exitSetCoordinate(self, ctx:mintParser.SetCoordinateContext):
        pass


    # Enter a parse tree produced by mintParser#orientation.
    def enterOrientation(self, ctx:mintParser.OrientationContext):
        pass

    # Exit a parse tree produced by mintParser#orientation.
    def exitOrientation(self, ctx:mintParser.OrientationContext):
        pass



del mintParser


from enum import Enum

from parchmint.layer import Layer


class MINTLayerType(Enum):
    FLOW = 1
    CONTROL = 2
    INTEGRATION = 3

    def __str__(self) -> str:
        if self == MINTLayerType.FLOW:
            return "FLOW"
        elif self == MINTLayerType.CONTROL:
            return "CONTROL"
        elif self == MINTLayerType.INTEGRATION:
            return "INTEGRATION"
        else:
            raise Exception("Could not generate MINT Layer string")

    def __eq__(self, o: object) -> bool:
        if o.__class__ is MINTLayerType:
            return super().__eq__(o)
        elif o.__class__ is str:
            if self is MINTLayerType.FLOW and o == "FLOW":
                return True
            elif self is MINTLayerType.CONTROL and o == "CONTROL":
                return True
            elif self is MINTLayerType.INTEGRATION and o == "INTEGRATION":
                return True
            else:
                return False
        else:
            return False


class MINTLayer(Layer):
    """MINT Layer class that abstracts the Layer class and adds additional
    methods to help generating MINT
    """

    def __init__(
        self,
        id: str,
        name: str,
        group: str,
        layer_type: MINTLayerType,
    ) -> None:
        """Creates a MINT layer object

        Args:
            id (str): unique id of the layer
            name (str): name of the layer
            group (str): group name
            layer_type (MINTLayerType): layer type of the layer
        """
        super(MINTLayer, self).__init__()
        self.ID = id
        self.group = group
        self.name = name
        self.type = str(layer_type)

    def to_MINT(self, content: str) -> str:
        """Generates the MINT string for the layer

        Args:
            content (str): MINT content that needs to be wrapped into the layer MINT

        Returns:
            str: Returns the MINT string
        """
        ret = "LAYER {} \n\n{} \n\nEND LAYER".format(self.type, content)
        return ret



from typing import Union

from parchmint.layer import Layer
from pymint.mintcomponent import MINTComponent
from pymint.mintlayer import MINTLayer


class MINTNode(MINTComponent):
    """NODE object that is considered a waypoint in the design
    however this is modelled as a component in Parchmint v1.

    simplifies the creation process and generate all of the
    required MINT.

    """

    def __init__(self, name: str, layer: Union[MINTLayer, Layer]) -> None:
        """Creates a new NODE object

        Args:
            name (str): name of the node
            layer (str, optional): [description].
        """
        super(MINTNode, self).__init__(name, "NODE", {}, [layer])

    def to_MINT(self) -> str:
        """Returns the MINT for the NODE

        Returns:
            str: [description]
        """
        ret = "NODE {};".format(self.name)
        return ret



from typing import List

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent


class ArrayConstraint(LayoutConstraint):
    """Array Constraint represents a 1D / 2D array with fixed spacings
    in both dimensions
    """

    def __init__(
        self,
        components: List[MINTComponent],
        xdim: int = 1,
        ydim: int = 1,
        horizontal_spacing: float = None,
        vertical_spacing: float = None,
    ) -> None:
        """Creates an instance of the array Constraints

        Args:
            components (List[MINTComponent]): List of components that need be covered
            by the constraint
            xdim (int, optional): X dimension of the array. Defaults to 1.
            ydim (int, optional): Y dimension of the array. Defaults to 1.
            horizontal_spacing (float, optional): horizontal spacing between the
            components (x-dimension). Defaults to None.
            vertical_spacing (float, optional): vertical spacing between the components
            (y-dimension). Defaults to None.
        """
        super().__init__()
        self._components.extend(components)
        self.__is1D = False
        self._horizontal_spacing = horizontal_spacing
        self._vertical_spacing = vertical_spacing

        if xdim is None:
            self.__xdim = len(self._components)
        else:
            self.__xdim = xdim

        if ydim is None:
            self.__is1D = True
            self.__ydim = 1
        else:
            self.__ydim = ydim

    @property
    def is1D(self) -> bool:
        """Returns true if 1D (BANK)

        Returns:
            bool: true if BANK
        """
        return self.__is1D

    @property
    def xdim(self) -> int:
        """Returns the x size of the array

        Returns:
            int: size
        """
        return self.__xdim

    @property
    def ydim(self) -> int:
        """Returns the y size of the array

        Returns:
            int: size
        """
        return self.__ydim

    @property
    def dim(self) -> int:
        """Returns the size of the array

        Returns:
            int: size
        """
        return self.__xdim



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from parchmint.component import Component
from parchmint.layer import Layer

if TYPE_CHECKING:
    from pymint import MINTLayer

from pymint.mintparams import MINTParams


class MINTComponent(Component):
    """Component Class abstracting parchmint component
    while adding helper methods to generate MINT

    """

    def __init__(
        self,
        name: str,
        technology: str,
        params: dict,
        layers: List[Union[MINTLayer, Layer]] = None,
    ) -> None:
        """Creates a MINT component

        Args:
            name (str): [description]
            technology (str): [description]
            params (dict): [description]
            layers (List[MINTLayer], optional): [description]. Defaults to None.

        Raises:
            Exception: hrows the exception if no layer information is present
        """
        super(MINTComponent, self).__init__()
        self.name = name
        self.ID = name
        self.entity = technology
        self.params = MINTParams(params)
        if layers is None:
            raise Exception("Cannot have no layer information")
        for layer in layers:
            self.layers.append(layer)

    def overwrite_id(self, id: str) -> None:
        """Overwites the ID

        Args:
            id (str): New id of the component
        """
        self.ID = id

    @property
    def component_spacing(self) -> int:
        """Returns the component spacing of the component

        Returns:
            int: Component Spacing Value in microns
        """
        return self.params.get_param("component_spacing")

    @component_spacing.setter
    def connection_spacing(self, value: int) -> None:
        """Sets the component spacing of the component

        Args:
            value (int): Value for the component spacing in microns
        """
        self.params.set_param("component_spacing", value)

    def to_MINT(self) -> str:
        """Returns the MINT String for the component

        Returns:
            str: This is the MINT string for the serialization
        """
        ret = "{} {} {};".format(self.entity, self.name, self.params.to_MINT())
        return ret

    def __hash__(self) -> int:
        return hash((self.ID, self.entity))



from typing import Union
from parchmint.layer import Layer
from pymint.mintcomponent import MINTComponent
from pymint.mintlayer import MINTLayer


class MINTTerminal(MINTComponent):
    """Class for representing the MINT terminal, its modelled as a
    Parchmint Component while giving helpful methods to model the design

    """

    def __init__(
        self, name: str, port_number: int, layer: Union[MINTLayer, Layer]
    ) -> None:
        super(MINTTerminal, self).__init__(name, "TERMINAL", {}, [layer])
        self.__port_number = port_number

    @property
    def port_number(self) -> int:
        """Returns the port number of the terminal (at a device level)

        Returns:
            int: port value
        """
        return self.__port_number

    def to_MINT(self) -> str:
        return "TERMINAL {} {}".format(self.name, self.__port_number)



from typing import List, Union

from parchmint.layer import Layer

from pymint.mintcomponent import MINTComponent
from pymint.mintlayer import MINTLayer


class MINTVia(MINTComponent):
    """Class to represent MINT Vias that are modelled as a Parchmint component.
    The class has helpful methods for adding and modifying the vias

    """

    def __init__(self, name: str, layers: List[Union[MINTLayer, Layer]]) -> None:
        """Creates a new instance of a Via

        Args:
            name (str): name/ID of the via
            layers (List[MINTLayer]): List of layer objects where the via is present
        """
        width = 0
        super(MINTVia, self).__init__(name, "VIA", {"width": width}, layers)

    def set_width(self, width: int) -> None:
        """Sets the width of the via

        Args:
            width (int): via width
        """
        self.params.set_param("width", width)

    def to_MINT(self) -> str:
        """Returns the MINT string of the via

        Returns:
            str: returns the via
        """
        return "VIA {}".format(self.name)



from enum import Enum
from typing import List

from pymint.mintcomponent import MINTComponent
from pymint.mintconnection import MINTConnection


class OperationType(Enum):
    """Enum that represents all the differnt types of constraints"""

    EXPLICIT_OPERATION = 0
    OPTIONAL_OPERATION = 1
    ALIGNMENT_OPERATION = 2
    SYMMETRY_OPERATION = 4
    RELATIVE_OPERATIONS = 5
    PARAMETER_MATCH = 6


class LayoutConstraint:
    """Base layout constraint class over which all the other constriaint
    classes will be based upon

    """

    def __init__(self) -> None:
        """Creates a new instance of the LayoutConstraint"""
        self._components = []
        self._connections = []
        self._type: OperationType

    def add_component(self, component: MINTComponent):
        """Adds a component to be covered by the layout constraint

        Args:
            component (MINTComponent): Component to be constrained
        """
        self._components.append(component)

    def add_connection(self, connection: MINTConnection):
        """Adds a connection to be covered by the layout constraint

        Args:
            connection (MINTConnection): Connection to be constrained
        """
        self._connections.append(connection)

    def get_components(self) -> List[MINTComponent]:
        """Returns components covered by the constraint

        Returns:
            List[MINTComponent]: Constrained components
        """
        return self._components

    def get_connections(self) -> List[MINTConnection]:
        """Returns the connections covered byt he constraint

        Returns:
            List[MINTConnection]: Constrained connections
        """
        return self._connections

    def contains_component(self, component: MINTComponent) -> bool:
        """Checks if the constraint covers the component

        Args:
            component (MINTComponent): component to check

        Returns:
            bool: true if component is present
        """
        return component in self._components

    def contains_connection(self, connection: MINTConnection) -> bool:
        """Checks if the constraint covers the connection

        Args:
            connection (MINTConnection): connection to check

        Returns:
            bool: true if connection is present
        """
        return connection in self._connections



from __future__ import annotations

import sys
from typing import List, Union
from parchmint.component import Component
from parchmint.connection import Connection

from parchmint.device import Device, ValveType

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent
from pymint.mintconnection import MINTConnection
from pymint.mintlayer import MINTLayer, MINTLayerType
from pymint.minttarget import MINTTarget
from pymint.mintterminal import MINTTerminal
from pymint.mintvia import MINTVia


class MINTDevice(Device):
    """Device class abstracting parchmint Device object adding additional
    methods for generating MINT

    """

    def __init__(self, name: str, json_data=None) -> None:
        """Creates a MINT device

        A MINT device has the extra bells and whistles necessary for genrating
        structured MINT. Pass the ParchMINT JSON to the `json` arg if you want
        to initialize from a JSON file.

        Args:
            name (str): name of the device
            json (str, optional): JSON string. Defaults to None.
        """
        super(MINTDevice, self).__init__(json_data=json_data)

        self.name = name
        self._layout_constraints = []
        self._terminals = []
        self._vias = []

    def create_mint_component(
        self, name: str, technology: str, params: dict, layer_ids: List[str]
    ) -> MINTComponent:
        """Creates a new component and adds it to the device

        Args:
            name (str): name
            technology (str): MINT string
            params (dict): params dictionary
            layer_ids (List[str]): list of layer ids

        Returns:
            MINTComponent: the newly created component
        """
        # Retrieve the correct layer:
        layers = []
        for layer_id in layer_ids:
            layers.append(super().get_layer(layer_id))
        component = MINTComponent(name, technology, params, layers)
        super().add_component(component)
        return component

    def create_mint_connection(
        self,
        name: str,
        technology: str,
        params: dict,
        source: MINTTarget,
        sinks: List[MINTTarget],
        layer_id: str,
    ) -> MINTConnection:
        """Creates a new MINT connection and adds it to the device

        Args:
            name (str): name of the connection
            technology (str): MINT string
            params (dict): dictionary of the paraeters
            source (MINTTarget): object defining where the connection starts
            sinks (List[MINTTarget]): list of objects defining where the connection ends
            layer_id (str): layer id of the connection

        Raises:
            Exception: Throws an exception if layer with the given id is not found

        Returns:
            MINTConnection: Returns the newly created connection
        """
        layer = self.get_layer(layer_id)
        if layer is None:
            raise Exception("Cannot create new MINT connection with invalid layer")
        if layer is None:
            raise Exception("Layer is None")
        connection = MINTConnection(name, technology, params, source, sinks, layer)
        self.add_connection(connection)
        return connection

    def create_mint_layer(
        self, id: str, name_postfix: str, group, layer_type: MINTLayerType
    ) -> MINTLayer:
        """[summary]

        Args:
            id (str): [description]
            name_postfix (str): [description]
            group ([type]): [description]
            layer_type (MINTLayerType): [description]

        Returns:
            MINTLayer: [description]
        """
        name = "{}_{}".format(str(MINTLayerType.FLOW), name_postfix)
        layer = MINTLayer(id, name, group, layer_type)
        super().add_layer(layer)
        return layer

    def create_valve(
        self,
        name: str,
        technology: str,
        params: dict,
        layer_ids: List[str],
        connection: Union[MINTConnection, Connection],
        valve_type: ValveType = ValveType.NORMALLY_OPEN,
    ) -> MINTComponent:
        """Creates a new valve and adds it to the device

        Args:
            name (str): name of the valve
            technology (str): MINT string
            params (dict): dictionary of the paraeters
            layer_ids (List[str]): list of layer ids
            connection (Union[MINTConnection, Connection]): connection to attach the valve to
            valve_type (ValveType, optional): valve type of the valve . Defaults to ValveType.NORMALLY_OPEN.

        Raises:
            Exception: [description]

        Returns:
            MINTComponent: [description]
        """
        valve = self.create_mint_component(name, technology, params, layer_ids)
        if connection not in self.connections:
            raise Exception("Connection {} not found".format(connection.ID))
        self.map_valve(valve, connection, valve_type)
        return valve

    def get_constraints(self) -> List[LayoutConstraint]:
        """Returns the layout constraints of the device. Currently does not support the
        constraints

        Returns:
            List[LayoutConstraint]: List of layout constriants objects
        """
        return self._layout_constraints

    def add_constraint(self, constraint: LayoutConstraint) -> None:
        """Adds a layout constriant to the device

        Args:
            constraint (LayoutConstraint): new constraint we want to add
        """
        self._layout_constraints.append(constraint)

    def to_MINT(self) -> str:
        """Returns the MINT string of the device

        Returns:
            str: MINT string
        """
        # TODO: Eventually I need to modify the MINT generation to account for all the
        # layout constraints

        # Generate a valve list for the device
        valve_list = self.valves
        full_layer_text = ""
        # Loop Over all the layers
        for layer in self.layers:
            componenttext = "\n".join(
                [
                    item.to_MINT()
                    for item in self.components
                    if item.layers[0] == layer and item not in valve_list
                ]
            )

            valvetext = ""
            if layer.type is str(MINTLayerType.CONTROL):
                valvetext = "\n".join(
                    [
                        MINTDevice.to_valve_MINT(item, self.get_valve_connection(item))
                        for item in valve_list
                        if layer in item.layers
                    ]
                )

            connectiontext = "\n".join(
                [item.to_MINT() for item in self.connections if item.layer == layer]
            )

            full_layer_text += (
                layer.to_MINT(
                    "{}\n\n{}\n\n{}".format(componenttext, valvetext, connectiontext)
                )
                + "\n\n"
            )

        full = "DEVICE {}\n\n{}".format(self.name, full_layer_text)
        return full

    def add_terminal(self, name: str, pin_number: int, layer_id: str) -> MINTTerminal:
        """Creates and adds a terminal to the device with an associated pin number

        Args:
            name (str): name of the pin
            pin_number (int): pin number of the terminal
            layer (str): layerid associated with the terminal

        Returns:
            MINTTerminal: The newly created terminal
        """
        layer = self.get_layer(layer_id)
        ret = MINTTerminal(name, pin_number, layer)
        self._terminals.append(ret)
        self.components.append(ret)
        return ret

    def add_via(self, name: str, layers: List[MINTLayer]) -> MINTVia:
        """Creates and adds a via to the device

        Args:
            name (str): name of the via
            width (int): width of the via
            layers (List[MINTLayer]): layers associated with the via

        Returns:
            MINTVia: The newly created via
        """
        ret = MINTVia(name, layers)
        self._vias.append(ret)
        self.components.append(ret)
        return ret

    def to_parchmint_v1(self) -> dict:
        """Returns the Parchmint Version 1.0 in the form of a dictionary

        Returns:
            dict: dictionary representing the json datastructure
        """
        ret = {}
        ret["name"] = self.name
        ret["components"] = [c.to_parchmint_v1() for c in self.components]
        ret["connections"] = [c.to_parchmint_v1() for c in self.connections]
        ret["params"] = self.params.to_parchmint_v1()
        ret["layers"] = [layer.to_parchmint_v1() for layer in self.layers]
        ret["version"] = 1

        return ret

    @staticmethod
    def to_valve_MINT(
        component: Union[MINTComponent, Component],
        connection: Union[MINTConnection, Connection],
    ) -> str:
        """Returns the MINT string of a valve

        This functions returns the MINT string of a valve.

        Args:
            component (Union[MINTComponent, Component]): [description]
            connection (Union[MINTConnection, Connection]): [description]

        Returns:
            str: [description]
        """
        return "{} {} on {} {};".format(
            component.entity, component.ID, connection.ID, component.params.to_MINT()
        )

    @staticmethod
    def from_mint_file(filepath: str, skip_constraints: bool = False) -> MINTDevice:
        """Compiles the MINT file at the given path

        Args:
            filepath (str): absolute filepath of the mint file
            skip_constraints (bool, optional): flag to accept / skip constraint parsing.
            Defaults to False.

        Returns:
            MINTDevice: The parsed device from the MINT file
        """
        import io

        from antlr4 import CommonTokenStream, FileStream, ParseTreeWalker

        from pymint.antlrgen.mintLexer import mintLexer
        from pymint.antlrgen.mintParser import mintParser
        from pymint.constraints.constraintlistener import ConstraintListener
        from pymint.mintcompiler import MINTCompiler
        from pymint.mintErrorListener import MINTErrorListener

        finput = FileStream(filepath)

        lexer = mintLexer(finput)

        stream = CommonTokenStream(lexer)

        parser = mintParser(stream)

        # Connect the Error Listener
        parse_output = io.StringIO()
        parse_output.write("MINT SYNTAX ERRORS:\n")

        error_listener = MINTErrorListener(parse_output)
        parser.addErrorListener(error_listener)

        tree = parser.netlist()

        if error_listener.pass_through is False:
            print("STOPPED: Syntax Error(s) Found")
            sys.exit(0)

        walker = ParseTreeWalker()

        listener = MINTCompiler()

        walker.walk(listener, tree)

        if listener.current_device is None:
            raise AssertionError
        current_device = listener.current_device

        if skip_constraints is not True:
            print("Computing Constraints")
            constraint_listener = ConstraintListener(listener.current_device)

            walker.walk(constraint_listener, tree)

            current_device = listener.current_device

        return current_device



from typing import List

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintconnection import MINTConnection


class LengthConstraint(LayoutConstraint):
    """Layout Constriant that will be used for specifying a fixed length
    for a connection

    """

    def __init__(self, connection: MINTConnection, length: float) -> None:
        """Creates a length constraint

        Args:
            component (MINTComponent): component covered by the constraint
            length (float): length of the connection that needs to be fixed
        """
        super().__init__()
        self._connections = []
        self._connections.append(connection)
        self.__length = length

    @property
    def connections(self) -> List[MINTConnection]:
        """Returns the component covered by the constraint

        Returns:
            MINTComponent: constrained component
        """
        return self._connections

    @property
    def length(self) -> float:
        """Returns the length fixed by the constraint

        Returns:
            float: length constrained
        """
        return self.__length



from typing import Optional

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent


class PositionConstraint(LayoutConstraint):
    """Layout constraint that fixes the location"""

    def __init__(
        self,
        component: MINTComponent,
        xpos: Optional[int],
        ypos: Optional[int],
        zpos: Optional[int],
    ) -> None:
        """Returns the position constrinant

        Args:
            component (MINTComponent): component covered by the constraint
            xpos (Optional[int]): x position value
            ypos (Optional[int]): y position value
            zpos (Optional[int]): z position value
        """
        super().__init__()
        self.add_component(component)
        self.__xpos = xpos
        self.__ypos = ypos
        self.__zpos = zpos

    def get_component(self) -> MINTComponent:
        """Returns the component constrained by the component

        Raises:
            Exception: if the component is not covered by the constraint

        Returns:
            MINTComponent: returns the component covered by the constraint
        """
        ret = self._components[0]
        if ret is None:
            raise Exception("No component set for PositionConstraint")

        return ret

    @property
    def xpos(self) -> Optional[int]:
        """Returns the x position of the component

        Returns:
            Optional[int]: x coordinate of the component if set else returns None
        """
        return self.__xpos

    @property
    def ypos(self) -> Optional[int]:
        """Returns the y position of the component

        Returns:
            Optional[int]: y coordinate of the component if set else returns None
        """
        return self.__ypos

    @property
    def zpos(self) -> Optional[int]:
        """Returns the z position of the component

        Returns:
            Optional[int]: z coordinate of the component if set else returns None
        """
        return self.__zpos



from enum import Enum
from typing import Dict

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent


class ComponentOrientation(Enum):
    """Enum ot represent component's relative orientation"""

    HORIZONTAL = 0
    VERTICAL = 1


class OrientationConstraint(LayoutConstraint):
    """Layout constraint that setups relative orientations for a
    number of components inside a single layer

    """

    def __init__(self) -> None:
        """Creates an instance of the orientation constraint"""
        super().__init__()
        self.__orientation_map: Dict[MINTComponent, ComponentOrientation] = {}

    def add_component_orientation_pair(
        self, component: MINTComponent, orientation: ComponentOrientation
    ) -> None:
        """Adds a component onto the constraint

        Args:
            component (MINTComponent): component covered by the constraint
            orientation (ComponentOrientation): orientation to set
        """
        super().add_component(component)
        self.__orientation_map[component] = orientation

    @property
    def orientation_map(self) -> Dict[MINTComponent, ComponentOrientation]:
        """Retuns the map of all the components and their corresponding orientation

        Returns:
            Dict[MINTComponent, ComponentOrientation]: dict mapping the components and
             their orientations
        """
        return self.__orientation_map



from pymint import MINTComponent
from pymint.constraints.constraint import LayoutConstraint


class RotationConstraint(LayoutConstraint):
    """Layout constraint that fixed the absolute rotation
    of the component

    """

    def __init__(self, component: MINTComponent, rotation: float) -> None:
        """Creates a Rotation constraint

        Args:
            component (MINTComponent): Component to be covered by the constraint
            rotation (float): rotation fixed by the constraint
        """
        super().__init__()
        self._components.append(component)
        self.__rotation = rotation

    @property
    def component(self) -> MINTComponent:
        """Returns the component covered by the constraint

        Returns:
            MINTComponent: constrianed component
        """
        return self._components[0]

    @property
    def rotation(self) -> float:
        """Returns the rotation fixed by the rotation

        Returns:
            float: rotation value
        """
        return self.__rotation



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from parchmint.connection import Connection
from parchmint.layer import Layer
from parchmint.target import Target

if TYPE_CHECKING:
    from pymint import MINTLayer

from pymint.mintparams import MINTParams
from pymint.minttarget import MINTTarget


class MINTConnection(Connection):
    """Connection class abstracting parchmint Connection class that adding
    helper methods to generate MINT

    """

    def __init__(
        self,
        name: str,
        technology: str,
        params: dict,
        source: Union[MINTTarget, Target],
        sinks: List[MINTTarget],
        layer: Union[MINTLayer, Layer],
    ) -> None:
        """Creates a new connection

        Args:
            name (str): name of the connection
            technology (str): MINT string
            params (dict): parameters
            source (MINTTarget): where the connection starts
            sinks (List[MINTTarget]): where the connection ends
            layer (MINTLayer, optional): layer information. Defaults to None.
        """
        super(MINTConnection, self).__init__()
        self.name = name
        self.ID = name
        self.entity = technology
        self.params: MINTParams = MINTParams(params)
        self.source: Union[MINTTarget, Target] = source
        self.sinks: List[MINTTarget] = sinks
        self.layer = layer

    def overwrite_id(self, id: str) -> None:
        """Overwites the ID

        Args:
            id (str): New id of the connection
        """
        self.ID = id

    @property
    def connection_spacing(self) -> int:
        """Returns the connection spacing of the component

        Returns:
            int: connection Spacing Value in microns
        """
        return self.params.get_param("channel_spacing")

    @connection_spacing.setter
    def connection_spacing(self, value: int) -> None:
        """Sets the component spacing of the connection

        Args:
            value (int): Value for the connection spacing in microns
        """
        self.params.set_param("channel_spacing", value)

    def to_MINT(self) -> str:
        """Returns the MINT String for the connection

        Returns:
            str: This is the MINT string for the serialization
        """
        ret = "{} {} from {} to {} {} ;".format(
            self.entity,
            self.name,
            self.source.to_MINT(),
            ", ".join([item.to_MINT() for item in self.sinks]),
            self.params.to_MINT(),
        )
        return ret



from .mintcompiler import MINTCompiler
from .mintcomponent import MINTComponent
from .mintconnection import MINTConnection
from .mintdevice import MINTDevice
from .mintlayer import MINTLayer, MINTLayerType
from .mintnode import MINTNode
from .mintparams import MINTParams
from .minttarget import MINTTarget
from .mintterminal import MINTTerminal
from .mintvia import MINTVia



from typing import List

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent
from pymint.mintdevice import MINTDevice


class MirrorConstraint(LayoutConstraint):
    """Layout constraint when two differnt sub-netlists need to have
    the same layout
    """

    def __init__(self, source_component: MINTComponent, mirror_count=None):
        """Create a new instance of the mirror constraint

        Args:
            source_component (MINTComponent): source for the mirror component to search
            for mirror groups
            mirror_count ([type], optional): number of mirror groups. Defaults to None.
        """
        super().__init__()
        self.__mirror_source: MINTComponent = source_component
        self.__mirror_groups: List[List[MINTComponent]] = []
        self.__mirror_count: int = mirror_count

    def add_group(self, components: List[MINTComponent]) -> None:
        """Adds the passed componets to a new group

        Args:
            components (List[MINTComponent]): List of components that need to be in a
            mirror group
        """
        self.__mirror_groups.append(components)

    @property
    def mirror_source(self) -> MINTComponent:
        """Returns the mirror source component

        Returns:
            MINTComponent: mirror source
        """
        return self.__mirror_source

    @property
    def mirror_groups(self) -> List[List[MINTComponent]]:
        """Returns the mirror groups

        Returns:
            List[List[MINTComponent]]: Mirror groups covered by the constraint
        """
        return self.__mirror_groups

    def find_mirror_candidates(self, device: MINTDevice) -> None:
        """Traverses the device on the mirror groups

        Args:
            device (MINTDevice): device to traverse

        Raises:
            Exception: if component is not part of any group
        """
        forward_traverse = True
        # TODO -
        # Step 1 - Go through all the outgoing components and create create mirror
        # groups
        G = device.G
        mirror_groups = []

        outgoing_edges = list(G.out_edges(self.__mirror_source.ID))
        incoming_edges = list(G.in_edges(self.__mirror_source.ID))
        edges_to_use = None
        if len(outgoing_edges) == self.__mirror_count:
            forward_traverse = True
            edges_to_use = outgoing_edges
        elif len(incoming_edges) == self.__mirror_count:
            forward_traverse = False
            edges_to_use = incoming_edges
        else:
            # Reverse Case Considtion
            if len(incoming_edges) == self.__mirror_count:
                forward_traverse = True
                edges_to_use = outgoing_edges
            elif len(outgoing_edges) == self.__mirror_count:
                forward_traverse = False
                edges_to_use = incoming_edges
            else:
                raise Exception(
                    "Unable to compute mirror group for source: {}, please check if"
                    " outgoing and incoming channels are declared correctly".format(
                        self.__mirror_source
                    )
                )

        sources = []

        for i in range(self.__mirror_count):
            if forward_traverse is True:
                component = device.get_component(edges_to_use[i][1])
            else:
                component = device.get_component(edges_to_use[i][0])
            mirror_groups.append([component])
            sources.append(component.ID)

        if forward_traverse is True:
            self.step_forward(sources, mirror_groups, device)
        else:
            self.step_reverse(sources, mirror_groups, device)

        # Step 2 - Save teh mirror groups into our constraint object
        self.__mirror_groups = mirror_groups

    def step_forward(
        self,
        sources: List[str],
        mirror_groups: List[List[MINTComponent]],
        device: MINTDevice,
    ) -> bool:
        """Steps in the forward direction

        Args:
            sources (List[str]): [description]
            mirror_groups (List[List[MINTComponent]]): [description]
            device (MINTDevice): [description]

        Raises:
            Exception: [description]

        Returns:
            bool: [description]
        """
        G = device.G
        outgoing_edges = []
        for source in sources:
            outgoing_edges.extend(list(G.out_edges(source)))

        sink_ids = []
        for edge in list(outgoing_edges):
            sink_ids.append(edge[1])

        # If there is a only 1 common sink, or less than the mirror count, then we
        # gotto kill the mirror groupings
        if len(sink_ids) < self.__mirror_count:
            return False

        # Check if all the types of the components are the same
        entities = []
        components = []
        for id in sink_ids:
            component = device.get_component(id)
            entities.append(component.entity)
            components.append(component)

        # If all the entities are the same, include them into the mirror groups
        entity_0 = entities[0]
        fail_flag = False
        for i, unused_element in enumerate(sink_ids):
            entity_i = entities[i]
            if entity_i != entity_0:
                fail_flag = True

        if fail_flag is True:
            return False

        # Since it works, we add everything to the mirror groups, for each of the
        # components see which group the predecessor is in and place it in the
        # corresponding group, if its not in any of the groups something
        # went wrong in the alg
        for component in components:
            assing_group_found_flag = False
            for edge in list(G.in_edges(component.ID)):
                predecessor_component = device.get_component(edge[0])
                for group in mirror_groups:
                    if predecessor_component in group:
                        group.append(component)
                        assing_group_found_flag = True
                        break

            if assing_group_found_flag is False:
                raise Exception(
                    "Could not find the mirror group for component: {}".format(
                        component.ID
                    )
                )

        next_sources = [c.ID for c in components]
        # self.step_forward(next_sources, mirror_groups, device)

        return True

    def step_reverse(
        self,
        sources: List[str],
        mirror_groups: List[List[MINTComponent]],
        device: MINTDevice,
    ) -> bool:
        """Takes a step in reverse


        Args:
            sources (List[str]): [description]
            mirror_groups (List[List[MINTComponent]]): [description]
            device (MINTDevice): [description]

        Raises:
            Exception: [description]

        Returns:
            bool: [description]
        """
        G = device.G
        incoming_edges = []
        for source in sources:
            incoming_edges.extend(list(G.in_edges(source)))

        sink_ids = []
        for edge in list(incoming_edges):
            sink_ids.append(edge[0])

        # If there is a only 1 common sink, or less than the mirror count, then we
        # gotto kill the mirror groupings
        if len(sink_ids) < self.__mirror_count:
            return False

        # Check if all the types of the components are the same
        entities = []
        components = []
        for id in sink_ids:
            component = device.get_component(id)
            entities.append(component.entity)
            components.append(component)

        # If all the entities are the same, include them into the mirror groups
        entity_0 = entities[0]
        fail_flag = False
        for i, unused_element in enumerate(sink_ids):
            entity_i = entities[i]
            if entity_i != entity_0:
                fail_flag = True

        if fail_flag is True:
            return False

        # Since it works, we add everything to the mirror groups, for each of the
        # components see which group the predecessor is in and place it in the
        # corresponding group, if its not in any of the groups something
        # went wrong in the alg
        for component in components:
            assing_group_found_flag = False
            for edge in list(G.in_edges(component.ID)):
                predecessor_component = device.get_component(edge[0])
                for group in mirror_groups:
                    if predecessor_component in group:
                        group.append(component)
                        assing_group_found_flag = True
                        break

            if assing_group_found_flag is False:
                raise Exception(
                    "Could not find the mirror group for component: {}".format(
                        component.ID
                    )
                )

        next_sources = [c.ID for c in components]
        # self.step_forward(next_sources, mirror_groups, device)

        return True



from typing import List

import networkx as nx

from pymint.constraints.constraint import LayoutConstraint
from pymint.mintcomponent import MINTComponent
from pymint.mintdevice import MINTDevice


class OrthogonalConstraint(LayoutConstraint):
    """Layout constraint used to represent an orthogonal layout
    requirement for a sub-netlist

    """

    def __init__(self, components: List[MINTComponent]) -> None:
        """Creates an orthogonal constraint

        Args:
            components (List[MINTComponent]): components covered by the constraint
        """
        super().__init__()
        self._components.extend(components)

    @staticmethod
    def traverse_node_component_neighbours(
        component: MINTComponent, device: MINTDevice
    ) -> List[MINTComponent]:
        """Traverses the node networks to find the components covered by the constraint

        Args:
            component (MINTComponent): starting component
            device (MINTDevice): current device

        Returns:
            List[MINTComponent]: list of components covered by the constraint
        """
        current_device = device
        G = device.G
        nodes = []
        nodes.append(component)

        def dfs_traverse_nodes(root_node: str):
            """Does a DFS traversal based on the node

            (to be used as a recursive function)

            Args:
                root_node (str): node from which we need to do the traversal
            """
            neighbors = list(nx.neighbors(G, root_node))
            for neighbor in neighbors:
                component = current_device.get_component(neighbor)
                if component in nodes:
                    continue
                if component.entity == "NODE":
                    nodes.append(component)
                    dfs_traverse_nodes(neighbor)
                else:
                    nodes.append(component)

        dfs_traverse_nodes(component.ID)

        return nodes



from parchmint.params import Params


class MINTParams(Params):
    """Params class that abstracts Parchmint Params with
    MINT compatible options

    """

    def __init__(self, pairs: dict) -> None:
        """Creates a MINTParams object

        Args:
            pairs (dict): [description]
        """
        super(MINTParams, self).__init__(None)
        for key in pairs.keys():
            self.data[key] = pairs[key]

    def to_MINT(self) -> str:
        """Returns the MINT string for the set params, it skips the connection/constriant/position params

        Returns:
            str: MINT string
        """
        skip_list = ["paths", "wayPoints", "position"]
        ret = ""
        for key in self.data.keys():
            if key in skip_list:
                continue
            ret += "{}={} ".format(key, self.data[key])
        return ret



from typing import Optional

from parchmint.target import Target


class MINTTarget(Target):
    """Target class that abstracts Parchmint Target class"""

    def __init__(self, componentstring: str, portstring: Optional[str] = None) -> None:
        super(MINTTarget, self).__init__(None)
        self.component = componentstring
        self.port = portstring

    def to_MINT(self) -> str:
        """MINT formatted string of the target  <component_name, port>

        Returns:
            str: MINT string
        """
        ret = "{} {}".format(self.component, "" if self.port is None else self.port)
        return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymint', 'pymint.antlrgen', 'pymint.constraints']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.8,<5.0',
 'install>=1.3.3,<2.0.0',
 'parchmint>=0.2.11,<0.3.0',
 'pip>=20.2.2,<21.0.0']

entry_points = \
{'console_scripts': ['mint-tools = pymint.cmdline:main']}

setup_kwargs = {
    'name': 'pymint',
    'version': '0.2.16',
    'description': 'MINT is a human readable language to describe Microfluidic Hardware Netlists. MINT is the name of the Microfluidic Netlist language used to describe microfluidic devices for Fluigi to place and route. Mint is a flavor of (MHDL) Microfluidic Hardware Description Language that can be used to represent Microfluidic Circuits.',
    'long_description': None,
    'author': 'Radhakrishna Sanka',
    'author_email': 'rkrishnasanka@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from io import StringIO

from antlr4.error.ErrorListener import ErrorListener


class MINTErrorListener(ErrorListener):
    """Error listener that will fire all error messages

    Listens to the errors and formats the output in a user readable manner
    """

    def __init__(self, output: StringIO):
        """Constructor

        Args:
            output (StringIO): IO stream object where the output is supposed to be
            written to
        """
        super().__init__()
        self.output = output
        self._symbol = ""
        self.pass_through = True

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        """Writes the syntax error to the listener output

        Args:
            recognizer ([type]): ANTLR internal
            offendingSymbol ([type]): Symbol which caused the error
            line ([type]): Line number
            column ([type]): Column number
            msg ([type]): Message generated by ANTLR
            e ([type]): ANTLR internal
        """
        self.output.write("Line {}:{} - {}".format(line, column, msg))
        self._symbol = offendingSymbol.text
        self.pass_through = False

    @property
    def symbol(self):
        """Symbol associated with the current error

        Returns:
            [String]: Error symbol
        """
        return self._symbol



import logging
from typing import List

from pymint.antlrgen.mintListener import mintListener
from pymint.antlrgen.mintParser import mintParser
from pymint.constraints.arrayconstraint import ArrayConstraint
from pymint.constraints.mirrorconstraint import MirrorConstraint
from pymint.constraints.orientationconstraint import (
    ComponentOrientation,
    OrientationConstraint,
)
from pymint.constraints.orthogonalconstraint import OrthogonalConstraint
from pymint.constraints.positionconstraint import PositionConstraint
from pymint.mintcomponent import MINTComponent
from pymint.mintdevice import MINTDevice
import re


class ConstraintListener(mintListener):
    """Listener that is used to catch all the layout constratints
    from the MINT design

    """

    def __init__(self, device: MINTDevice):
        super().__init__()
        self.current_device = device

        # Temporary store for constrained components
        self._constrained_components = []

        # Temporary storage for mirrored components source
        self._mirror_constraints = []

        # Temporary store for position constraints
        self._xpos = None
        self._ypos = None
        self._zpos = None

        # Temporary store for array constraints
        self._vertical_spacing = None
        self._horizontal_spacing = None
        self._spacing = None

        # Temporary store for relative orientatino constraints
        self._orientation = None

        # Global Relative Orientation Constraints
        self._global_relative_operations: List[OrientationConstraint] = []

    def enterPositionConstraintStat(
        self, ctx: mintParser.PositionConstraintStatContext
    ):
        self._xpos = None
        self._ypos = None
        self._zpos = None

    def enterSetCoordinate(self, ctx: mintParser.SetCoordinateContext):
        coordinate_label = ctx.coordinate
        coordinate_value = int(ctx.INT().getText())

        if coordinate_label == "X":
            self._xpos = coordinate_value
        elif coordinate_label == "Y":
            self._ypos = coordinate_value
        elif coordinate_label == "Z":
            self._zpos = coordinate_value
        else:
            raise Exception("Invalid coordinate label !")

    def exitPositionConstraintStat(self, ctx: mintParser.PositionConstraintStatContext):
        constraint = PositionConstraint(
            self._constrained_components[0], self._xpos, self._ypos, self._zpos
        )
        self.current_device.add_constraint(constraint)

    def enterHorizontalSpacingParam(
        self, ctx: mintParser.HorizontalSpacingParamContext
    ):
        self._horizontal_spacing = float(ctx.value().getText())

    def enterVerticalSpacingParam(self, ctx: mintParser.VerticalSpacingParamContext):
        self._vertical_spacing = float(ctx.value().getText())

    def enterSpacingParam(self, ctx: mintParser.SpacingParamContext):
        self._spacing = float(ctx.value().getText())

    def exitGridStat(self, ctx: mintParser.GridStatContext):
        xdim = 1
        ydim = 1

        if ctx.xdim is not None:
            xdim = int(ctx.xdim.text)
        else:
            logging.warning(
                "No X Dimension found for GRID stat, setting dimension to 1"
            )
        if ctx.ydim is not None:
            ydim = int(ctx.ydim.text)
        else:
            logging.warning(
                "No Y Dimension found for GRID stat, setting dimension to 1"
            )
        # We need to add all the parameters here
        constraint = ArrayConstraint(
            self._constrained_components,
            xdim,
            ydim,
            self._horizontal_spacing,
            self._vertical_spacing,
        )

        self.current_device.add_constraint(constraint)

    def exitBankStat(self, ctx: mintParser.BankStatContext):
        dim = 1
        if ctx.dim is not None:
            dim = int(ctx.dim.text)
        else:
            logging.warning("No dimension found for BANK stat, setting dimension to 1")
        # We need to add all the parameters here
        constraint = ArrayConstraint(
            self._constrained_components, dim, horizontal_spacing=self._spacing
        )

        self.current_device.add_constraint(constraint)

    def enterOrientation(self, ctx: mintParser.OrientationContext):
        if ctx.getText() == "H":
            self._orientation = ComponentOrientation.HORIZONTAL
        elif ctx.getText() == "V":
            self._orientation = ComponentOrientation.VERTICAL
        else:
            logging.error("Orientation that is not H or V found. Illegal Syntax")
            raise Exception("Orientation that is not H or V found. Illegal Syntax")

    def enterUfname(self, ctx: mintParser.UfnameContext):
        element_name = ctx.getText()
        component = None
        connection = None
        if self.current_device.component_exists(element_name):
            component = self.current_device.get_component(element_name)
        elif self.current_device.connection_exists(element_name):
            connection = self.current_device.get_connection(element_name)
        if component is not None:
            self._constrained_components.append(component)
        elif connection is not None:
            self._constrained_components.append(connection)
        else:
            # Check if theres a regex match against all the component names/id's
            component_names = [
                component.ID for component in self.current_device.components
            ]
            # Check if theres a regex match against all the connection id's in component_name
            matches = map(
                lambda x: re.match(f"{element_name}_\\d+(_\\d+)?", x), component_names
            )
            for match in matches:
                if match is not None:
                    self._constrained_components.append(
                        self.current_device.get_component(match.group(0))
                    )
                    break
            else:
                print(
                    'Could not find component or connection with the ID "{}" in device'
                    .format(element_name)
                )
                raise Exception(
                    f"Component {element_name} not found while processing constraint"
                )

    def enterLayerBlock(self, ctx: mintParser.LayerBlockContext):
        # Create a new relative orientation constraint for the whole layer
        self._global_relative_operations.append(OrientationConstraint())

    def enterFlowStat(self, ctx: mintParser.FlowStatContext):
        self._constrained_components = []
        self._orientation = None

    def exitFlowStat(self, ctx: mintParser.FlowStatContext):
        # Skip if there's no orientation set
        if self._orientation is None:
            return

        # In general check whats there and set the constraint for all the items
        # in the statement
        constraint = self._global_relative_operations[-1]
        for component in self._constrained_components:
            constraint.add_component_orientation_pair(component, self._orientation)

        self.current_device.add_constraint(constraint)

    def exitNodeStat(self, ctx: mintParser.NodeStatContext):
        # TODO: Expand on neighbours until we hit all the components on the node
        # periphery
        for component in self._constrained_components:
            if component is None:
                raise Exception(
                    "Could not apply Orthogonal Constraint, {} component not found !"
                    .format(ctx.getText())
                )

            if self._checkIfComponentConstranied(component):
                continue

            # TODO: check if component exists in any of the of existing constraints
            components = OrthogonalConstraint.traverse_node_component_neighbours(
                component, self.current_device
            )
            constraint = OrthogonalConstraint(components)
            self.current_device.add_constraint(constraint)

        # TODO: Add all the components onto the list and create the constraint
        pass

    def exitChannelStat(self, ctx: mintParser.ChannelStatContext):
        # TODO - If length constraints exists, create them here
        pass

    def exitSpanStat(self, ctx: mintParser.SpanStatContext):
        for component in self._constrained_components:
            # Generate mirror constraints based on the in and out dimensins of span
            in_size = int(ctx.indim.text)
            out_size = int(ctx.outdim.text)

            if in_size != 1:
                constraint = MirrorConstraint(component, in_size)
                self._mirror_constraints.append(constraint)

            if out_size != 1:
                constraint = MirrorConstraint(component, out_size)
                self._mirror_constraints.append(constraint)

    def exitLayerBlock(self, ctx: mintParser.LayerBlockContext):
        for constraint in self._mirror_constraints:
            if not isinstance(constraint, MirrorConstraint):
                raise AssertionError
            try:
                constraint.find_mirror_candidates(self.current_device)
                self.current_device.add_constraint(constraint)
            except Exception as e:
                print(e)

    # ------------ Helpers ----------

    def _checkIfComponentConstranied(self, component: MINTComponent) -> bool:
        found_flag = False
        for constraint in self.current_device.get_constraints():
            if isinstance(constraint, OrthogonalConstraint):
                found_flag = constraint.contains_component(component)

        return found_flag



# Generated from ./mint.g4 by ANTLR 4.9.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .mintParser import mintParser
else:
    from mintParser import mintParser

# This class defines a complete generic visitor for a parse tree produced by mintParser.

class mintVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by mintParser#netlist.
    def visitNetlist(self, ctx:mintParser.NetlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#importBlock.
    def visitImportBlock(self, ctx:mintParser.ImportBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#importStat.
    def visitImportStat(self, ctx:mintParser.ImportStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#header.
    def visitHeader(self, ctx:mintParser.HeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufmoduleBlock.
    def visitUfmoduleBlock(self, ctx:mintParser.UfmoduleBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#globalStats.
    def visitGlobalStats(self, ctx:mintParser.GlobalStatsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufmoduleStat.
    def visitUfmoduleStat(self, ctx:mintParser.UfmoduleStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#layerBlocks.
    def visitLayerBlocks(self, ctx:mintParser.LayerBlocksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#layerBlock.
    def visitLayerBlock(self, ctx:mintParser.LayerBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#flowBlock.
    def visitFlowBlock(self, ctx:mintParser.FlowBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#controlBlock.
    def visitControlBlock(self, ctx:mintParser.ControlBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#integrationBlock.
    def visitIntegrationBlock(self, ctx:mintParser.IntegrationBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#flowStat.
    def visitFlowStat(self, ctx:mintParser.FlowStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#controlStat.
    def visitControlStat(self, ctx:mintParser.ControlStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#integrationStat.
    def visitIntegrationStat(self, ctx:mintParser.IntegrationStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#primitiveStat.
    def visitPrimitiveStat(self, ctx:mintParser.PrimitiveStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#bankDeclStat.
    def visitBankDeclStat(self, ctx:mintParser.BankDeclStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#bankGenStat.
    def visitBankGenStat(self, ctx:mintParser.BankGenStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#bankStat.
    def visitBankStat(self, ctx:mintParser.BankStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#gridGenStat.
    def visitGridGenStat(self, ctx:mintParser.GridGenStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#gridDeclStat.
    def visitGridDeclStat(self, ctx:mintParser.GridDeclStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#gridStat.
    def visitGridStat(self, ctx:mintParser.GridStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#spanStat.
    def visitSpanStat(self, ctx:mintParser.SpanStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#valveStat.
    def visitValveStat(self, ctx:mintParser.ValveStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#nodeStat.
    def visitNodeStat(self, ctx:mintParser.NodeStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#viaStat.
    def visitViaStat(self, ctx:mintParser.ViaStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#terminalStat.
    def visitTerminalStat(self, ctx:mintParser.TerminalStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#channelStat.
    def visitChannelStat(self, ctx:mintParser.ChannelStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#netStat.
    def visitNetStat(self, ctx:mintParser.NetStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#entity.
    def visitEntity(self, ctx:mintParser.EntityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#entity_element.
    def visitEntity_element(self, ctx:mintParser.Entity_elementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#paramsStat.
    def visitParamsStat(self, ctx:mintParser.ParamsStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#statTerminaion.
    def visitStatTerminaion(self, ctx:mintParser.StatTerminaionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#connectionParamStat.
    def visitConnectionParamStat(self, ctx:mintParser.ConnectionParamStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#paramStat.
    def visitParamStat(self, ctx:mintParser.ParamStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#constraintParams.
    def visitConstraintParams(self, ctx:mintParser.ConstraintParamsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#spacingParam.
    def visitSpacingParam(self, ctx:mintParser.SpacingParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#directionParam.
    def visitDirectionParam(self, ctx:mintParser.DirectionParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#param_element.
    def visitParam_element(self, ctx:mintParser.Param_elementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#intParam.
    def visitIntParam(self, ctx:mintParser.IntParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#boolParam.
    def visitBoolParam(self, ctx:mintParser.BoolParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#widthParam.
    def visitWidthParam(self, ctx:mintParser.WidthParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#verticalSpacingParam.
    def visitVerticalSpacingParam(self, ctx:mintParser.VerticalSpacingParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#horizontalSpacingParam.
    def visitHorizontalSpacingParam(self, ctx:mintParser.HorizontalSpacingParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#rotationParam.
    def visitRotationParam(self, ctx:mintParser.RotationParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#lengthParam.
    def visitLengthParam(self, ctx:mintParser.LengthParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufmodulename.
    def visitUfmodulename(self, ctx:mintParser.UfmodulenameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufterminal.
    def visitUfterminal(self, ctx:mintParser.UfterminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#uftargets.
    def visitUftargets(self, ctx:mintParser.UftargetsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#uftarget.
    def visitUftarget(self, ctx:mintParser.UftargetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufname.
    def visitUfname(self, ctx:mintParser.UfnameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#ufnames.
    def visitUfnames(self, ctx:mintParser.UfnamesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#value.
    def visitValue(self, ctx:mintParser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#boolvalue.
    def visitBoolvalue(self, ctx:mintParser.BoolvalueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#positionConstraintStat.
    def visitPositionConstraintStat(self, ctx:mintParser.PositionConstraintStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#setCoordinate.
    def visitSetCoordinate(self, ctx:mintParser.SetCoordinateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by mintParser#orientation.
    def visitOrientation(self, ctx:mintParser.OrientationContext):
        return self.visitChildren(ctx)



del mintParser